@@ -156,6 +156,24 @@ inline ZWord pop(ZStack* stack)
   return *(--stack->stack_top);
 }
 
+inline ZWord top(ZStack* stack)
+{
+    if (stack->current_frame)
+    {
+#ifdef SAFE
+        if (stack->current_frame->frame_size <= 0)
+            zmachine_fatal("Stack underflow");
+#endif
+    }
+
+#ifdef SAFE
+    if (stack->stack_top == stack->stack)
+        zmachine_fatal("Stack underflow");
+#endif
+
+    return *(stack->stack_top-1);
+}
+
 ZFrame* call_routine(ZDWord* pc, ZStack* stack, ZDWord start)
 {
   ZFrame* newframe;
@@ -169,6 +187,7 @@ ZFrame* call_routine(ZDWord* pc, ZStack* stack, ZDWord start)
   newframe->storevar     = 0;
   newframe->discard      = 0;
   newframe->frame_size   = 0;
+  newframe->break_on_return = 0;
   if (stack->current_frame != NULL)
     newframe->frame_num  = stack->current_frame->frame_num+1;
   else
@@ -231,13 +250,39 @@ inline void store(ZStack* stack, int var, ZWord value)
     }
 }
 
+inline void store_nopush(ZStack* stack, int var, ZWord value)
+{
+#ifdef DEBUG
+    printf_debug("Storing %i in Variable #%x\n", value, var);
+#endif
+    if (var == 0)
+    {
+#ifdef SAFE
+        if (stack->stack_top == stack->stack)
+            zmachine_fatal("Stack underflow");
+#endif
+
+        *(stack->stack_top-1) = value;
+    }
+    else if (var < 16)
+    {
+        stack->current_frame->local[var] = value;
+    }
+    else
+    {
+        var-=16;
+        machine.globals[var<<1]     = value>>8;
+        machine.globals[(var<<1)+1] = value;
+    }
+}
+
 void restart_machine(void)
 {
   machine.screen_on = 1;
   machine.transcript_on = 0;
   if (machine.transcript_file)
     {
-      fclose(machine.transcript_file);
+      close_file(machine.transcript_file);
       machine.transcript_file = NULL;
     }
   machine.memory_on = 0;
@@ -249,6 +294,7 @@ void restart_machine(void)
     }
 
   zmachine_setup_header();
+  display_has_restarted();
 }
 
 #define Obj3(x) (machine.memory + GetWord(machine.header, ZH_objs) + 62+(((x)-1)*9))
@@ -447,22 +493,6 @@ static void tracking_print(char* format, ...)
 }
 #endif
 
-static void zcode_op_print_obj_123(ZStack* stack, ZWord arg)
-{
-  ZByte* obj;
-  ZByte* prop;
-  int len;
-
-  obj = Obj3(arg);
-  prop = machine.memory + ((obj[7]<<8)|obj[8]) + 1;
-
-#ifdef DEBUG
-  /* printf_debug(">%s<\n", zscii_to_ascii(prop, &len)); */
-#endif
-
-  stream_prints(zscii_to_unicode(prop, &len));
-}
-
 #ifdef SUPPORT_VERSION_3
 static void draw_statusbar_123(ZStack* stack)
 {
@@ -485,7 +515,7 @@ static void draw_statusbar_123(ZStack* stack)
   display_erase_line(1);
   display_set_cursor(2, 0);
 
-  display_prints(zscii_to_unicode(prop, &len));
+  display_prints((int*)zscii_to_unicode(prop, &len));
 
   score = GetVar(17);
   moves = GetVar(18);
@@ -584,6 +614,7 @@ inline static int convert_colour(int col)
 char save_fname[256] = "savefile.qut";
 char script_fname[256] = "script.txt";
 
+#if WINDOW_SYSTEM != 3 && WINDOW_SYSTEM !=4
 static void get_fname(char* name, int len, int save)
 {
 #if WINDOW_SYSTEM != 2
@@ -693,10 +724,12 @@ static void get_fname(char* name, int len, int save)
     }
 #endif
 }
+#endif
 
-#if WINDOW_SYSTEM != 3
+#if WINDOW_SYSTEM != 3 && WINDOW_SYSTEM != 4
 ZFile* get_file_write(int* fsize,
-		      char* save_fname)
+		      char* save_fname,
+                      ZFile_type purpose)
 {
   int fs, ok;
 
@@ -715,7 +748,7 @@ ZFile* get_file_write(int* fsize,
 	stream_printf("That file already exists!\nAre you sure? (y/N) ");
 	stream_readline(yn, 1, 0);
 	
-	if (tolower(yn[0]) == 'y')
+	if (unicode_to_lower(yn[0]) == 'y')
 	  ok = 1;
 	else
 	  {
@@ -732,7 +765,8 @@ ZFile* get_file_write(int* fsize,
 }
 
 ZFile* get_file_read(int* fsize,
-		     char* save_fname)
+		     char* save_fname,
+                     ZFile_type purpose)
 {
   int fs;
 
@@ -756,7 +790,7 @@ static int save_1234(ZDWord  pc,
   ZFile* f;
 
   stream_printf("\nPlease supply a filename for save\n");
-  f = get_file_write(NULL, save_fname);
+  f = get_file_write(NULL, save_fname, ZFile_save);
   
   if (st >= 0)
     store(stack, st, 2);
@@ -784,7 +818,7 @@ static int restore_1234(ZDWord* pc, ZStack* stack)
   ZDWord sz;
 
   stream_printf("\nPlease supply a filename for restore\n");
-  f = get_file_read(&sz, save_fname);
+  f = get_file_read(&sz, save_fname, ZFile_save);
   
   if (state_load(f, sz, stack, pc))
     {
@@ -824,54 +858,21 @@ static void zcode_op_output_stream(ZStack* stack,
     case 2:
       if (machine.transcript_file == NULL)
 	{
-	  int fname[256];
-	  int x;
-
-	  for (x=0; script_fname[x] != 0; x++)
-	    fname[x] = script_fname[x];
-	  fname[x] = 0;
-	  stream_printf("\nPlease supply a filename for transcript\nFile: ");
-	  stream_readline(fname, 256, 0);
-	  for (x=0; fname[x] != 0; x++)
-	    script_fname[x] = fname[x];
-	  script_fname[x] = 0;
-
-	  if (get_file_size(script_fname) != -1)
-	    {
-	      int yn[5];
-
-	      yn[0] = 0;
-	      stream_printf("That file already exists!\nAre you sure? (y/N) ");
-	      stream_readline(yn, 1, 0);
-	      
-	      if (tolower(yn[0]) == 'y')
-		machine.transcript_file = fopen(script_fname, "a");
-	    }
-	  else
-	    machine.transcript_file = fopen(script_fname, "a");
-	  if (machine.transcript_file == NULL)
-	    {
-	      stream_printf("Couldn't open %s for writing.\n", script_fname);
-	      return;
-	    }
-	  else
-	    {
-	      machine.transcript_on = 1;
-	      fprintf(machine.transcript_file, "*** Transcript generated by Zoom\n\n");
-	    }
+	  machine.transcript_file = get_file_write(NULL, script_fname, ZFile_transcript);
 	}
 
-      if (machine.transcript_file != NULL)
+      if (machine.transcript_file != NULL) {
+	write_stringf(machine.transcript_file, "*** Transcript generated by Zoom\n\n");
+
 	machine.transcript_on = 1;
 
-      w = Word(ZH_flags2);
-      w |= 1;
-      machine.memory[ZH_flags2] = w>>8;
-      machine.memory[ZH_flags2+1] = w;
+	w = Word(ZH_flags2);
+	w |= 1;
+	machine.memory[ZH_flags2] = w>>8;
+	machine.memory[ZH_flags2+1] = w;
+      }
       break;
     case -2:
-      if (machine.transcript_file != NULL)
-	fflush(machine.transcript_file);
       machine.transcript_on = 0;
 
       w = Word(ZH_flags2);
@@ -927,40 +928,11 @@ static void zcode_op_output_stream(ZStack* stack,
     case 4:
       if (machine.transcript_file == NULL)
 	{
-	  int fname[256];
-	  int x;
-
-	  for (x=0; script_fname[x] != 0; x++)
-	    fname[x] = script_fname[x];
-	  fname[x] = 0;
-	  stream_printf("\nPlease supply a filename for transcript\nFile: ");
-	  stream_readline(fname, 256, 0);
-	  for (x=0; fname[x] != 0; x++)
-	    script_fname[x] = fname[x];
-	  script_fname[x] = 0;
-
-	  if (get_file_size(script_fname) != -1)
-	    {
-	      int yn[5];
-
-	      yn[0] = 0;
-	      stream_printf("That file already exists!\nAre you sure? (y/N) ");
-	      stream_readline(yn, 1, 0);
-	      
-	      if (tolower(yn[0]) == 'y')
-		machine.transcript_file = fopen(script_fname, "a");
-	    }
-	  else
-	    machine.transcript_file = fopen(script_fname, "a");
-	  if (machine.transcript_file == NULL)
-	    {
-	      stream_printf("Couldn't open %s for writing.\n", script_fname);
-	      return;
-	    }
-	  else
-	    {
-	      machine.transcript_commands = 1;
-	    }
+	  machine.transcript_file = get_file_write(NULL, script_fname, ZFile_recording);
+	
+	  if (machine.transcript_file) {
+	    machine.transcript_commands = 1;
+	  }
 	}
 
       if (machine.transcript_file != NULL)
@@ -969,8 +941,6 @@ static void zcode_op_output_stream(ZStack* stack,
       
       break;
     case -4:
-      if (machine.transcript_file != NULL)
-	fflush(machine.transcript_file);
       machine.transcript_commands = 0;
       break;
       
@@ -1051,7 +1021,7 @@ static void zcode_op_aread_5678(ZDWord* pc,
 				int st)
 {
   ZByte* mem;
-  int* buf;
+  unsigned int* buf;
   int x;
   
   mem = machine.memory + (ZUWord) args->arg[0];
@@ -1084,7 +1054,7 @@ static void zcode_op_aread_5678(ZDWord* pc,
 	}
       buf[x] = 0;
 
-      stream_remove_buffer(buf);
+      stream_remove_buffer((int*)buf);
     }
   else
     buf[0] = 0;
@@ -1093,7 +1063,7 @@ static void zcode_op_aread_5678(ZDWord* pc,
 
   if (Word(ZH_termtable) != 0)
     {
-      char* table = NULL;
+      unsigned char* table = NULL;
       int pos, tablelen;
 
       tablelen = 0;
@@ -1121,7 +1091,7 @@ static void zcode_op_aread_5678(ZDWord* pc,
     {
       int res;
       
-      res = stream_readline(buf, mem[0], 0);
+      res = stream_readline((int*)buf, mem[0], 0);
       display_terminating(NULL);
 
       if (res == 254 || res == 253)
@@ -1154,7 +1124,7 @@ static void zcode_op_aread_5678(ZDWord* pc,
     {
       int res;
 
-      res = stream_readline(buf, mem[0], args->arg[2]*100);
+      res = stream_readline((int*)buf, mem[0], args->arg[2]*100);
       display_terminating(NULL);
       store(stack, st, res);
       
@@ -1167,7 +1137,7 @@ static void zcode_op_aread_5678(ZDWord* pc,
 	  for (x=0; buf[x] != 0; x++)
 	    {
 	      mem[1]++;
-	      buf[x] = tolower(buf[x]);
+	      buf[x] = unicode_to_lower(buf[x]);
 	      mem[x+2] = zscii_get_char(buf[x]);
 	    }
 
@@ -1213,7 +1183,7 @@ static void zcode_op_aread_5678(ZDWord* pc,
   for (x=0; buf[x] != 0; x++)
     {
       mem[1]++;
-      buf[x] = tolower(buf[x]);
+      buf[x] = unicode_to_lower(buf[x]);
       mem[x+2] = zscii_get_char(buf[x]);
     }
 
@@ -1323,8 +1293,8 @@ static void zcode_op_sread_4(ZDWord* pc,
 
 	  for (x=0; buf[x] != 0; x++)
 	    {
-	      buf[x] = tolower(buf[x]);
-	      mem[x+1] = tolower(x);
+	      buf[x] = unicode_to_lower(buf[x]);
+	      mem[x+1] = zscii_get_char(buf[x]);
 	    }
 	  mem[x+1] = 0;
 
@@ -1340,14 +1310,14 @@ static void zcode_op_sread_4(ZDWord* pc,
 
   for (x=0; buf[x] != 0; x++)
     {
-      buf[x] = tolower(buf[x]);
+      buf[x] = unicode_to_lower(buf[x]);
       mem[x+1] = zscii_get_char(buf[x]);
     }
   mem[x+1] = 0;
 
   if (args->n_args > 1)
     {
-      tokenise_string(buf,
+      tokenise_string((int*)buf,
 		      Word(ZH_dict),
 		      machine.memory + (ZUWord) args->arg[1],
 		      0,
@@ -1737,6 +1707,42 @@ void zmachine_runsome(const int version,
   int *              string;
 
   int x;
+  
+  /*
+   * PowerMac Dual 1.25Ghz:
+   *
+   * Without computed gotos:
+   * ZMark version 0.2, by Andrew Hunter
+   * 
+   * IntMark1: 2.18 secs
+   * IntMark2: 2.29 secs
+   * JumpMark: 2.96 secs
+   * CallMark: 0.38 secs
+   * NopMark: 1.09 secs
+   * 
+   * With computed gotos:
+   * 
+   * ZMark version 0.2, by Andrew Hunter
+   * 
+   * IntMark1: 2.23 secs
+   * IntMark2: 2.30 secs
+   * JumpMark: 3.16 secs
+   * CallMark: 0.40 secs
+   * NopMark: 1.32 secs
+   *
+   * (Hmm, gcc 3.3 is a major improvement over 2.95: NopMark used to be 5x slower with computed gotos)
+   *
+   * Bizarrely, speed seems to have improved now I've added the sigusr1 stuff
+   */
+  
+  /*
+   * The 'computed gotos' style interpreter is faster to compile, but can 
+   * run a bit slower under certain circumstances. Instead of switch statements,
+   * we use tables of gotos and then a large list of labels. This is non-portable
+   * to other compilers (but Zoom won't compile without gcc at the moment anyway).
+   *
+   * I'm going to look into improving this
+   */
 
 #ifdef HAVE_COMPUTED_GOTOS
 # define TABLES_ONLY
@@ -1754,7 +1760,7 @@ void zmachine_runsome(const int version,
 
   pc = start_counter;
   stack = &machine.stack;
-
+	  
 #ifdef HAVE_COMPUTED_GOTOS
   switch (version)
     {
@@ -1799,6 +1805,14 @@ void zmachine_runsome(const int version,
     }
 
  loop:
+#ifdef REMOTE_BREAKPOINT
+	  /* Really need to find a way to do this without a performance impact */
+	  if (machine.force_breakpoint) {
+		  machine.force_breakpoint = 0;
+		  debug_set_breakpoint(pc, 1, 0);
+	  }
+#endif
+	  
   instr = GetCode(pc);
  execute_instr:
   goto *decode[instr];
@@ -1826,8 +1840,16 @@ void zmachine_runsome(const int version,
 # endif
 
 #else
-  while (1)
+  for(;;)
     {
+#ifdef REMOTE_BREAKPOINT
+	  /* Really need to find a way to do this without a performance impact */
+	  if (machine.force_breakpoint) {
+		  machine.force_breakpoint = 0;
+		  debug_set_breakpoint(pc, 1, 0);
+	  }
+#endif
+
       instr = GetCode(pc);
 
 #ifdef DEBUG
