@@ -130,8 +130,9 @@ NSString* ZoomSkeinItemPboardType = @"ZoomSkeinItemPboardType";
 	
 	if (trackingRects) [trackingRects release];
 
-	if (itemToEdit) [itemToEdit release];
+	if (itemToEdit)    [itemToEdit release];
 	if (fieldScroller) [fieldScroller release];
+	if (fieldStorage)  [fieldStorage release];
 	
 	if (trackedItem)   [trackedItem release];
 	if (clickedItem)   [clickedItem release];
@@ -438,6 +439,9 @@ NSString* ZoomSkeinItemPboardType = @"ZoomSkeinItemPboardType";
 		while (item = [itemEnum nextObject]) {
 			NSRect itemRect = [layout activeAreaForItem: item];
 			
+			if (!NSIntersectsRect(itemRect, visibleRect)) continue;
+			itemRect = NSIntersectionRect(itemRect, visibleRect);
+			
 			// Same reasoning as before
 			[trackingItems addObject: item];
 			inside = NO;
@@ -858,13 +862,25 @@ NSString* ZoomSkeinItemPboardType = @"ZoomSkeinItemPboardType";
 	if (itemToEdit != nil && fieldEditor != nil) {
 		ZoomSkeinItem* parent = [itemToEdit parent];
 		
+		BOOL samename;		// Set to YES if we need to remove and re-add
+		
+		if (!editingAnnotation && [parent childWithCommand: [fieldEditor string]] != itemToEdit) 
+			samename = YES;
+		else
+			samename = NO;
+		
 		// This will merge trees if the item gets the same name as a neighbouring item
-		[itemToEdit removeFromParent];
+		if (samename) [itemToEdit removeFromParent];
 		if (!editingAnnotation)
 			[itemToEdit setCommand: [fieldEditor string]];
 		else
 			[itemToEdit setAnnotation: [fieldEditor string]];
-		ZoomSkeinItem* newItem = [parent addChild: itemToEdit];
+		ZoomSkeinItem* newItem;
+		
+		if (samename)
+			newItem = [parent addChild: itemToEdit];
+		else
+			newItem = itemToEdit;
 		
 		// Change the active item if required
 		if (itemToEdit == [skein activeItem]) {
@@ -942,7 +958,7 @@ NSString* ZoomSkeinItemPboardType = @"ZoomSkeinItemPboardType";
 	NSRect itemFrame = [layout textAreaForItem: skeinItem];
 	
 	// (Or the annotation)
-	if (annotation) itemFrame.origin.y -= 20;
+	if (annotation) itemFrame.origin.y -= 18;
 	
 	// Make sure the item is the right size
 	float minItemWidth = itemWidth - 32.0;
@@ -976,24 +992,30 @@ NSString* ZoomSkeinItemPboardType = @"ZoomSkeinItemPboardType";
 	fieldEditor = (NSTextView*)[[self window] fieldEditor: YES
 												forObject: self];
 	
+	if (fieldStorage) [fieldStorage release];
+	fieldStorage = [[NSTextStorage alloc] initWithString: itemText
+											  attributes: itemTextAttributes];	
+	[[fieldEditor textStorage] setAttributedString: fieldStorage];
+	[fieldEditor setSelectedRange: NSMakeRange(0,0)];
+	
 	[fieldEditor setDelegate: self];
 	[fieldScroller setFrame: itemFrame];
 	[fieldEditor setFrame: NSInsetRect(itemFrame, 2.0, 2.0)];
 	
-	[[fieldEditor textStorage] setAttributedString: [[[NSAttributedString alloc] initWithString: itemText
-																					 attributes: itemTextAttributes] autorelease]];
 	[fieldEditor setAlignment: NSCenterTextAlignment];
+	[fieldEditor setFont: [itemTextAttributes objectForKey: NSFontAttributeName]];
 	
 	[fieldEditor setRichText:NO];
 	[fieldEditor setAllowsDocumentBackgroundColorChange:NO];
-	[fieldEditor setBackgroundColor:[NSColor clearColor]];
+	[fieldEditor setBackgroundColor:[NSColor whiteColor]];
 	
 	[[fieldEditor textContainer] setContainerSize: NSMakeSize(NSInsetRect(itemFrame, 2.0, 2.0).size.width, 1e6)];
 	[[fieldEditor textContainer] setWidthTracksTextView:NO];
 	[[fieldEditor textContainer] setHeightTracksTextView:NO];
 	[fieldEditor setHorizontallyResizable:NO];
 	[fieldEditor setVerticallyResizable:YES];
-	[fieldEditor setDrawsBackground: NO];
+	[fieldEditor setDrawsBackground: YES];
+	[fieldEditor setEditable: YES];
 	
 	// Activate it
 	[fieldScroller setDocumentView: fieldEditor];
@@ -1027,6 +1049,11 @@ NSString* ZoomSkeinItemPboardType = @"ZoomSkeinItemPboardType";
 	return [layout selectedItem];
 }
 
+- (void) highlightSkeinLine: (ZoomSkeinItem*) itemOnLine {
+	[layout highlightSkeinLine: itemOnLine];
+	[self setNeedsDisplay: YES];
+}
+
 // = Playing the game =
 
 - (void) playToPoint: (ZoomSkeinItem*) item {
